(**************************************************************************)
(* Sample code to show how comprehesions works                            *)
(*                                                                        *)
(*  Author(s):  Roberto Di Cosmo (2012 - )                                *)
(*                                                                        *)
(*  This is free software: you can redistribute it and/or modify          *)
(*  it under the terms of the GNU Lesser General Public License as        *)
(*  published by the Free Software Foundation, either version 2 of the    *)
(*  License, or (at your option) any later version.  A special linking    *)
(*  exception to the GNU Lesser General Public License applies to this    *)
(*  library, see the LICENSE file for more information.                   *)
(**************************************************************************)

#load "dynlink.cma";;
#load "camlp4o.cma";;
#load "Camlp4Parsers/Camlp4ListComprehension.cmo";;

let rec range i j = if i > j then [] else i :: range (i+1) j;;

(* tail recursive version of range *)

let range i j = let rec aux acc = function n -> if n < i then acc else aux (n::acc) (n-1) in aux [] j;;

let triples n = 
 [ (a,b,c) | a <- range 1 n; b <- range 1 n; c <- range 1 n; 
             a*a + b*b = c*c ];;

triples 10;;

let triples_ordered n = 
 [ (a,b,c) | a <- range 1 n; b <- range a n; c <- range b n; 
             a*a + b*b = c*c ];;

triples_ordered 10;;

let partition n = [ (a,b) | a <- range 1 n; b <- range 1 n;
                            a+b=n; a<=b];;
partition 10;;

let partition_better n = 
 [ (a,b) | a <- range 1 n; b <- range a n; a+b=n];;

let partition_efficient n = 
 [ (a,b) | a <- range 1 (n/2); b <- [n-a]];;

(* the expansion of partition *)

let partition n =
 List.concat
  (List.map
    (fun a ->
      List.map (fun b -> (a, b))
       (List.filter (fun b -> a <= b)
         (List.filter (fun b -> (a + b) = n) (range 1 n))))
    (range 1 n))
;;

let rec concatMap f = function 
 | [] -> [] 
 | a::r -> (f a) @ (concatMap f r);;

(* or equivalently *)
let concatMap f l = 
 List.fold_right (fun x acc -> (f x)@acc) l [] ;;

(* the List monad ... *)

let bind l f = concatMap f l
let return = fun x -> [x]
let zero = []
;;

module ListMonad = 
  struct
    type 'a t = 'a list
    let return x = [x]
    let bind l f = List.fold_right (fun x acc -> (f x)@acc) l []
    let zero = []
    let ( >>= ) l f  = bind l f
  end;;

open ListMonad;;

(* rewrite everything using just the monad combinators *)

(* warm up *)

(* [ (x,y) | x <- range 1 3; y <- range 1 3 ] *)

[1;2;3] >>= fun x -> 
[1;2;3] >>= fun y -> 
return (x,y);;

(* [ (x,y) | x <- range 1 3; 
             y <-- range 1 3; 
             x+y > 3 ] *)

[1;2;3] >>= fun x -> 
[1;2;3] >>= fun y -> 
if x+y > 3 then return (x,y) else zero;;

(* Pythagorean triples :
   [ (a,b,c) | a <- range 1 n; 
               b <- range a n; 
               c <- range b n; 
               a*a + b*b = c*c ] *)

let triples n = 
 range 1 n >>= fun a -> 
 range a n >>= fun b ->
 range b n >>= fun c -> 
 if a*a + b*b = c*c then return (a,b,c) else zero;;

(* [ (a,b) | a <- range 1 n;
             b <- range a n; 
             a+b=n] *)

let partition n = 
 range 1 n >>= fun a ->
 range a n >>= fun b ->
 if a+b=n then return (a,b) else zero;;

(* the Array monad *)

module ArrayMonad = 
  struct
    type 'a t = 'a array
    let return x = Array.create 1 x
    let bind a f = Array.fold_right 
        (fun x acc -> Array.concat [(f x); acc]) a [||]
    let zero = [| |]
    let ( >>= ) a f  = bind a f
  end

let range i j = Array.init (j-i+1) (fun x -> x+i);;

open ArrayMonad;;

(* warm up *)

(* [ (x,y) | x <- range 1 3; y <- range 1 3 ] *)

[|1;2;3|] >>= fun x -> 
[|1;2;3|] >>= fun y -> 
return (x,y);;

(* [ (x,y) | x <- range 1 3; 
             y <-- range 1 3; 
             x+y > 3 ] *)


[|1;2;3|] >>= fun x -> 
[|1;2;3|] >>= fun y -> 
if x+y > 3 then return (x,y) else zero;;

(* Pythagorean triples :
   [ (a,b,c) | a <- range 1 n; 
               b <- range a n; 
               c <- range b n; 
               a*a + b*b = c*c ] *)

let triples n = 
 range 1 n >>= fun a -> 
 range a n >>= fun b ->
 range b n >>= fun c -> 
 if a*a + b*b = c*c then return (a,b,c) else zero;;

(* [ (a,b) | a <- range 1 n;
             b <- range a n; 
             a+b=n] *)

let partition n = 
 range 1 n >>= fun a ->
 range a n >>= fun b ->
 if a+b=n then return (a,b) else zero;;


(* Now check the monad equations *)

(* bind m return = m *)
[|1;2;3|] >>= fun x -> return x;;

(* bind (return e) f = f e *)

return 3 >>= fun x -> return (2*x);;
(fun x -> return (2*x)) 3;;

(* bind e1 (fun x -> bind e2 (fun y -> e3)) 
   = bind (bind e1 (fun x -> e2)) (fun y -> e3) 
   si x non libre dans e3
 *)
[|1;2;3|] >>= fun x -> 
range x 3 >>= fun y -> return y;;

([|1;2;3|] >>= fun x -> range x 3)
>>= fun y -> return y;;

(* bind zero f = zero *)

zero >>= (fun x -> return (x+1));;
